/*
 * This file is part of the Disco Deterministic Network Calculator v2.1.0 "Hydra".
 *
 * Copyright (C) 2013 - 2015 Steffen Bondorf
 *
 * disco | Distributed Computer Systems Lab
 * University of Kaiserslautern, Germany
 *
 * http://disco.cs.uni-kl.de
 *
 *
 * The Disco Deterministic Network Calculator (DiscoDNC) is free software;
 * you can redistribute it and/or modify it under the terms of the 
 * GNU Lesser General Public License as published by the Free Software Foundation; 
 * either version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 *
 */

package unikl.disco.nc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import unikl.disco.curves.MaxServiceCurve;
import unikl.disco.curves.ServiceCurve;
import unikl.disco.curves.ArrivalCurve;
import unikl.disco.misc.SetUtils;
import unikl.disco.nc.AnalysisConfig;
import unikl.disco.nc.ArrivalBound;
import unikl.disco.nc.LeftOverService;
import unikl.disco.nc.AnalysisConfig.MuxDiscipline;
import unikl.disco.nc.AnalysisConfig.GammaFlag;
import unikl.disco.network.Flow;
import unikl.disco.network.Link;
import unikl.disco.network.Network;
import unikl.disco.network.Path;
import unikl.disco.network.Server;
import unikl.disco.network.Server.Multiplexing;
import unikl.disco.minplus.Convolution;
import unikl.disco.minplus.Deconvolution;

/**
 * 
 * @author Steffen Bondorf
 *
 */
public class PbooArrivalBound_Concatenation extends ArrivalBound {
	
	@SuppressWarnings("unused")
	private PbooArrivalBound_Concatenation() {}
	
	public PbooArrivalBound_Concatenation( Network network, AnalysisConfig configuration ) {
		this.network = network;
		this.configuration = configuration;
	}

	public Set<ArrivalCurve> computeArrivalBound( Link link, Flow flow_of_interest ) throws Exception {
		return computeArrivalBound( link, network.getFlows( link ), flow_of_interest );
	}

	public Set<ArrivalCurve> computeArrivalBound( Link link, Set<Flow> f_xfcaller, Flow flow_of_interest ) throws Exception {
		Set<ArrivalCurve> alphas_xfcaller = new HashSet<ArrivalCurve>( Collections.singleton( ArrivalCurve.createNullArrival() ) );
		if ( f_xfcaller.isEmpty() || f_xfcaller == null )
		{
			return alphas_xfcaller;
		}
		
		// Get the servers on common sub-path of f_xfcaller flows crossing link
		Server to = link.getDest();
		Set<Flow> f_to = network.getFlows( to );
		Set<Flow> f_xfcaller_to = SetUtils.getIntersection( f_to, f_xfcaller );
		f_xfcaller_to.remove( flow_of_interest );
		if ( f_xfcaller_to.size() == 0 )
		{
			return alphas_xfcaller;
		}

		// The shortcut found in PmooArrivalBound for the case that (from == to) will not be implemented here.
		// There's not a big potential to increase performance as the PMOO arrival bound implicitly handles this situation by only iterating over one server in the for loop.
		Server from = network.findSplittingServer( to, f_xfcaller_to );
		Flow f_representative = f_xfcaller_to.iterator().next();
		Path common_subpath = f_representative.getSubPath( from, link.getSource() );
		
		// Calculate the left-over service curves on this sub-path by convolution of the individual left over service curves
		Set<ServiceCurve> betas_lo_subpath = new HashSet<ServiceCurve>();
		for ( Server s : common_subpath.getServers() )
		{
			Set<ServiceCurve> betas_lo_s = new HashSet<ServiceCurve>();
			
			Set<Flow> f_xxfcaller_s = SetUtils.getDifference( network.getFlows( s ), f_xfcaller );
			f_xxfcaller_s.remove( flow_of_interest );

			if( f_xxfcaller_s.isEmpty() ) {
				betas_lo_s.add( s.getServiceCurve() );
				betas_lo_subpath = Convolution.convolve_SCs_SCs( betas_lo_subpath, betas_lo_s );
				continue;
			}
			
			Set<ArrivalCurve> alpha_xxfcaller_s = ArrivalBound.computeArrivalBounds( network, configuration, s, f_xxfcaller_s, flow_of_interest );
			
			// Calculate the left-over service curve for this single server
			if( configuration.multiplexingDiscipline() == MuxDiscipline.GLOBAL_FIFO
					|| ( configuration.multiplexingDiscipline() == MuxDiscipline.SERVER_LOCAL & s.multiplexingDiscipline() == Multiplexing.FIFO ) )
			{
				betas_lo_s = LeftOverService.fifoMux( s.getServiceCurve(), alpha_xxfcaller_s );
			}
			else
			{
				betas_lo_s = LeftOverService.arbMux( s.getServiceCurve(), alpha_xxfcaller_s );				
			}
			
			// Check if there's any service left on this path. If not, the set only contains a null-service curve.
			if( betas_lo_s.size() == 1 && betas_lo_s.iterator().next().equals( ServiceCurve.createNullService() ) ) {
				throw new Exception( "No service left over during PBOO arrival bounding!" );
			}
			
			// Combine into the sub-path's left-over service curve
			betas_lo_subpath = Convolution.convolve_SCs_SCs( betas_lo_subpath, betas_lo_s );
		}
		
		// Next we need to know the arrival bound of f_xfcaller at the server 'from', i.e., at the above sub-path's source
		// in order to deconvolve it with beta_lo_s to get the arrival bound of the sub-path
		// Note that flows f_xfcaller that originate in 'from' are covered by this call of computeArrivalBound
		Set<ArrivalCurve> alpha_xfcaller_from = ArrivalBound.computeArrivalBounds( network, configuration, from, f_xfcaller, flow_of_interest );

		if( configuration.useGamma() != GammaFlag.GLOBALLY_OFF  )
		{
			MaxServiceCurve gamma = common_subpath.getGamma();
			alphas_xfcaller = Deconvolution.deconvolve_almostConcCs_SCs( Convolution.convolve_ACs_MSC( alpha_xfcaller_from, gamma ), betas_lo_subpath );
		}
		else
		{
			alphas_xfcaller = Deconvolution.deconvolve( alpha_xfcaller_from, betas_lo_subpath );
		}
		
		if( configuration.useExtraGamma() != GammaFlag.GLOBALLY_OFF )
		{
			MaxServiceCurve extra_gamma = common_subpath.getExtraGamma();
			alphas_xfcaller = Convolution.convolve_ACs_EGamma( alphas_xfcaller, extra_gamma );
		}
		
		return alphas_xfcaller;
	}
}
